
#include <boost/algorithm/string/replace.hpp>
#include <boost/lexical_cast.hpp>

#include <iostream>
#include <vector>
#include <string>

#include <cgicc/Cgicc.h>
#include <cgicc/HTTPHTMLHeader.h>
#include <cgicc/HTMLClasses.h>

#include <libxml++/parsers/domparser.h>
#include <libxml++/document.h>

#include <log4cxx/logger.h>
#include <log4cxx/propertyconfigurator.h>
#include <log4cxx/helpers/exception.h>

#include "base/exec.h"
#include "util/enforcer.h"
#include "util/log.h"
#include "util/path.h"

#include <fstream>
#include <iostream>
#include <boost/iostreams/filtering_stream.hpp>
#include <boost/iostreams/filtering_streambuf.hpp>
#include <boost/iostreams/copy.hpp>
#include <boost/iostreams/filter/zlib.hpp>
#include <boost/thread/thread.hpp>
#include <boost/thread/mutex.hpp>
#include <boost/thread/condition.hpp>

using namespace std;
using namespace util;
using namespace cgicc;
using namespace xmlpp;
using namespace boost;
using namespace boost::iostreams;

using namespace log4cxx;
using namespace log4cxx::helpers;

static bool finished = false;

struct space_writer 
{
  // The Non-Parsed-Headers version nph-mbrace.cgi makes sure the spaces are not buffered by Apache,
  // but send directly to the client. As a result, when the client breaks the connection and 
  // a space is written, nph-mbrace.cgi aborts with a SIGPIPE. We don't handle the signal,
  // just let the program abort. This is exactly what we want because the client is gone anyway and 
  // CGI is a single-shot application. Note: This does NOT work with FastCGI!!

  space_writer() { }

  void operator()()
  {
    cout << "X-mBrace-Spaces: "; // Custom header to append spaces to.

    while (!finished)
    {
      mutex::scoped_lock lock(mut);
      cond.timed_wait(lock, get_system_time() + posix_time::seconds(1));
      if (!finished)
        cout << ' ' << flush;
    }

    cout << endl;
  }

  mutex mut;
  condition cond;
};

const string default_message = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n \
                                <message><credentials><token>unknown</token><user>unknown</user><client>unknown</client></credentials><requests><unknown><parms/><response><status>failure</status><payload/><error><code>ERROR_CODE</code><usermsg>USER_MESSAGE</usermsg><techmsg>TECH_MESSAGE</techmsg><module>cgi</module><file>FILE</file><line>LINE</line></error></response></unknown></requests></message>";

void output_error(const string &code, const string &file, int line, const string &usr_msg, const string &tech_msg = "")
{
  const char *eol = "\r\n";
  string reply(default_message);
  algorithm::replace_all(reply, "ERROR_CODE", code);
  algorithm::replace_all(reply, "USER_MESSAGE", usr_msg);
  algorithm::replace_all(reply, "FILE", file);
  algorithm::replace_all(reply, "LINE", lexical_cast<string>(line));
  algorithm::replace_all(reply, "TECH_MESSAGE", tech_msg.empty() ? usr_msg : tech_msg);

  LOG4CXX_ERROR(logger, "Reply:\n" + reply);

  cout << "Content-Type: application/xml" << endl; // Needs to be first header.
  cout << "Content-Length: " << reply.length() << eol << eol;
  cout << reply;
}

void output_reply(const string &reply)
{
  const char *eol = "\r\n";
  cout << "Content-Type: application/xml" << endl; // Needs to be first header.
  cout << "Content-Length: " << reply.length() << eol << eol;
  cout << reply << flush;
}

  int 
main(int argc, 
    char **argv)
{
  try 
  {
    Cgicc cgi;

    PropertyConfigurator::configure(INSTALL_PATH "/etc/log.cfg");

    string env = cgi.getEnvironment().getRequestMethod();
    if (env != "POST")
    {
      output_error("BAD_CGI_REQUEST_METHOD", __FILE__, __LINE__, "Expected POST request");
      return 0;
    }

    string type = cgi.getEnvironment().getContentType();
    if (type != "application/octet-stream" && type != "application/xml")
    {
      output_error("BAD_CGI_CONTENT_TYPE", __FILE__, __LINE__, "Expected content-type application/octet-stream or application/xml");
      return 0;
    }

    string post_data = cgi.getEnvironment().getPostData();
    if (post_data.empty())
    {
      output_error("NO_CGI_POST_DATA", __FILE__, __LINE__, "No POST data");
      return 0;
    }

    stringstream post_stream;
    istringstream in_post_stream(post_data);
    filtering_streambuf<iostreams::input> in;
    if (type == "application/octet-stream")
      in.push(zlib_decompressor());
    in.push(in_post_stream);
    iostreams::copy(in, post_stream);

    LOG4CXX_DEBUG(logger, "Request (decompressed):");
    LOG4CXX_DEBUG(logger, post_stream.str());

    DomParser parser;
    parser.set_substitute_entities();
    Document *doc = 0;
    try
    {
      // parser.parse_stream(post_stream) is extremely slow for large data!! 
      // Use parse_memory() instead.
      parser.parse_memory(post_stream.str()); 
      doc = parser.get_document();
    }
    catch (const xmlpp::exception &e)
    {
      output_error("XML_PARSE_FAILURE", __FILE__, __LINE__, "Failed to parse XML request", e.what());
      return 0;
    }

    ENFORCE(doc);

    cout << "HTTP/1.1 200 OK" << endl;
    space_writer spacer;
    thread thrd(ref(spacer));

    try
    {
      cgi::executor exec;
      cgi::executor::return_t result = exec(doc);

      string reply = doc->write_to_string();

      if (exec.error()) {
        LOG4CXX_ERROR(logger, "Send reply:\n" + reply);
      }
      else {
        LOG4CXX_DEBUG(logger, "Send reply:\n" + reply);
      }

      // Stop and notify space writer.
      {
        mutex::scoped_lock lock(spacer.mut);
        finished = true;
        spacer.cond.notify_one();
      }

      // Join with space writer.
      thrd.join();

      output_reply(reply);
    }
    catch (const std::exception &e)
    {
      output_error("REQUEST_HANDLER_FAILURE", __FILE__, __LINE__, "Failed to handle request", e.what());
    }
  }
  catch(iostreams::zlib_error& e)
  {
    LOG4CXX_FATAL(logger, e.error());
    ENFORCE(false)("Failed to handle request due to zlib-error: ")(e.error());
  }
  catch(std::exception& e) 
  {
    LOG4CXX_FATAL(logger, e.what());
    ENFORCE(false)("Failed to handle request: ")(e.what());
  }

  return 0;
}

