
#include "xml/getattr.h"
#include "xml/gettext.h"
#include "xml/find.h"
#include "util/log.h"
#include "base/factory.h"
#include "base/base.h"
#include "base/enforcer.h"
#include "mva/mva_exact.h"
#include "mva/mva_approx.h"
#include "mysql/dm/setting.h"

#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/thread.hpp>
#include <boost/bind.hpp>
#include <string>

#include <libxml++/nodes/element.h>
#include <libxml++/document.h>

using namespace xmlpp;
using namespace std;
using namespace boost;
using namespace boost::posix_time;
using namespace mva;
using namespace util;

namespace cgi { namespace {

  static const unsigned control_flow_indent = 3;
  static const char control_flow_char = '-';
  static const unsigned max_sqn_hqn_iterations = 30;

  struct sqn_mva_result
  {
    vector<double> _blocking_proc_vector;
    double _max_b;
    mva_job _sqn_job;
  };

  inline string number_to_string(const double num, const unsigned prec = 9)
  {
    stringstream str;
    str << fixed << setprecision(prec) << num;
    return str.str();
  }

  double hqn_residence_time(const mva_job & job, const string & class_id)
  {
    if (job.host_id() == 0)
      return .0;

    // We are NOT in initalization phase, so we have to correct service demands.
    const mva_class::ptrs::const_iterator cit = find_if(job.classes().begin(), job.classes().end(), 
        bind(&mva_class::id, _1) == class_id);
    CGI_ENFORCE(cit != job.classes().end(), "ERROR_INTERNAL_SERVER")
      ("Class ")(class_id)(" not found in hardware job.");

    return cit->residence_time();
  }

  double hqn_residence_time(const mva_job & hqn_job, 
      const string & class_id, 
      const string & device_id, 
      const mva_device::type_t device_type)
  {
    if (hqn_job.host_id() == 0)
      return .0;

    // We are NOT in initalization phase, so we have to correct service demands.
    const mva_class::ptrs::const_iterator hw_cit = find_if(hqn_job.classes().begin(), hqn_job.classes().end(), 
        bind(&mva_class::id, _1) == class_id);
    CGI_ENFORCE(hw_cit != hqn_job.classes().end(), "ERROR_INTERNAL_SERVER")
      ("Class ")(class_id)(" not found in hardware job.");

    const mva_device::ptrs::const_iterator hw_dit = find_if(hw_cit->devices().begin(), hw_cit->devices().end(), 
        bind(&mva_device::id, _1) == device_id && bind(&mva_device::type, _1) == device_type);
    /*
       if (hw_dit == hw_cit->devices().end())
       {
       cerr << "+++++++++++++++++++++++++++++++++\n";
       hqn_job.print_config(cerr);
       cerr << "+++++++++++++++++++++++++++++++++\n";
       }
     */
    CGI_ENFORCE(hw_dit != hw_cit->devices().end(), "ERROR_INTERNAL_SERVER")
      ("Device ")(device_id)(" not found in class ")(class_id)(" in hardware job.");

    return hw_dit->residence_time();
  }
}

template<typename S, typename C> 
class calc_mva_base : public base
{
  typedef C Calculator;
  typedef S Sub;

  public:

  calc_mva_base() : 
    base(), 
    _one_or_more_forced_tandem_breaks(false)
  { }
  virtual ~calc_mva_base() { }

  protected:

  virtual void do_exec();
  virtual bool authenticate(const credentials &cred) const { return true; } 
  virtual string model() const = 0;

  bool _one_or_more_forced_tandem_breaks;

  private:

  void kickstart_mva();

  mva_job do_mva(const unsigned host_id, 
      const vector<double> & num_customer_vector, 
      unsigned indent);
  mva_job do_tandem_mva(const unsigned host_id, 
      const vector<double> & num_customer_vector, 
      unsigned indent);
  mva_job do_hqn_mva(const unsigned host_id, 
      const vector<double> & num_customer_vector, 
      const std::vector<double> & blocking_proc_vector, 
      unsigned indent);
  sqn_mva_result do_sqn_mva_item(const Node * xml_softrc,
      const bool first_softrc,
      const unsigned local_host_id, 
      const mva_job & hqn_job, 
      const vector<double> & num_customer_vector, 
      const vector<double> & prev_blocking_proc_vector, 
      unsigned indent);
  sqn_mva_result do_sqn_mva(const unsigned local_host_id, 
      const mva_job & hqn_job, 
      const vector<double> & num_customer_vector, 
      const vector<double> & prev_blocking_proc_vector, 
      unsigned indent);
  mva_job do_client_mva(const unsigned client_host_id, const unsigned indent); // Separate post calculation with input from _parms.
  void do_root_mva(const unsigned root_host_id, const mva_job & main_job, const mva_job & client_mva, const unsigned indent); // Separate post calculation with input from do_mva().

  void generate_template_payload();
  void generate_payload(const mva_job & job, unsigned host_id);
  void generate_client_payload(const mva_job & job, unsigned host_id);
  void generate_root_payload(const mva_job & job, unsigned root_host_id);

  std::stringstream _debug_control_flow;
  std::stringstream _debug;
};

class calc_mva_exact : public calc_mva_base<calc_mva_exact, mva_exact>
{
  public:

    static std::string identifier() { return "calc-exact-mva"; }
    virtual string model() const { return "Closed Queuing with Exact MVA"; }
};

class calc_mva_approx : public calc_mva_base<calc_mva_approx, mva_approx>
{
  public:

    static std::string identifier() { return "calc-approx-mva"; }
    virtual string model() const { return "Closed Queuing with Approximate MVA"; }
};

  template<typename S, typename C>
void calc_mva_base<S, C>::kickstart_mva() 
{
  generate_template_payload();

  // Initialize the initial number of customer vector to values in the XML.
  unsigned class_index = 0;
  NodeSet xml_classes = _parms->find("classes/class");
  vector<double> num_customer_vector(xml_classes.size());
  for (NodeSet::const_iterator cit = xml_classes.begin();
      cit != xml_classes.end();
      ++cit, ++class_index) {
    num_customer_vector[class_index] = max<double>(1., get_attr_value<double>(*cit, "n"));
  }

  unsigned root_host_id = get_attr_value<unsigned>(_parms, "classes/class[1]/host[@type='root']", "id");
  mva_job main_job = do_mva(root_host_id, num_customer_vector, control_flow_indent);

  mva_job client_job;
  NodeSet xml_client_devices = _parms->find("classes/class[1]/host[@type='client']");
  if (!xml_client_devices.empty())
  {
    unsigned client_host_id = get_attr_value<unsigned>(xml_client_devices[0], "id");
    client_job = do_client_mva(client_host_id, control_flow_indent);
  }

  do_root_mva(root_host_id, main_job, client_job, control_flow_indent);
}

  template<typename S, typename C>
void calc_mva_base<S, C>::generate_template_payload()
{
  Element * xml_payload_classes = _payload->add_child("classes"); 
  const NodeSet xml_parms_classes = _parms->find("classes/class");
  for (NodeSet::const_iterator cit = xml_parms_classes.begin();
      cit != xml_parms_classes.end();
      ++cit)
  {
    const string class_id = get_attr_value<string>(*cit, "id");

    Element * xml_payload_class = xml_payload_classes->add_child("class");
    xml_payload_class->set_attribute("id", class_id);

    NodeSet xml_devicess = (*cit)->find("host");
    for (NodeSet::const_iterator dsit = xml_devicess.begin();
        dsit != xml_devicess.end();
        ++dsit)
    {
      const string host_id = get_attr_value<string>(*dsit, "id");
      Element * xml_payload_host = xml_payload_class->add_child("host");
      xml_payload_host->set_attribute("id", host_id);

      Element * xml_payload_device_cpu = xml_payload_host->add_child("device");
      xml_payload_device_cpu->set_attribute("id", "CPU");
      xml_payload_device_cpu->set_attribute("type", "loadindep");

      Element * xml_payload_device_io = xml_payload_host->add_child("device");
      xml_payload_device_io->set_attribute("id", "IO");
      xml_payload_device_io->set_attribute("type", "loadindep");

      NodeSet xml_devices = (*dsit)->find("device");
      for (NodeSet::const_iterator dit = xml_devices.begin();
          dit != xml_devices.end();
          ++dit)
      {
        const string device_id = get_attr_value(*dit, "id");
        if (device_id.substr(0, 3) != "NI-" && 
            device_id.substr(0, 3) != "NO-" && 
            device_id.substr(0, 3) != "NL-") 
          continue;

        Element * xml_payload_device_net = xml_payload_host->add_child("device");
        xml_payload_device_net->set_attribute("id", device_id);
        xml_payload_device_net->set_attribute("type", "loadindep");
      }

      NodeSet xml_softrcs = _parms->find("softrcs/softrc");
      for (NodeSet::const_iterator sit = xml_softrcs.begin();
          sit != xml_softrcs.end();
          ++sit)
      {
        const string softrc_host_id = get_attr_value<string>(*sit, "hostId");
        if (softrc_host_id != host_id)
          continue;

        const string softrc_id = get_attr_value<string>(*sit, "id");
        Element * xml_payload_device_softrc = xml_payload_host->add_child("device");
        xml_payload_device_softrc->set_attribute("id", softrc_id);
        xml_payload_device_softrc->set_attribute("type", "loadindep");
      }
    }
  }
}

  template<typename S, typename C>
void calc_mva_base<S, C>::generate_payload(const mva_job & job, unsigned host_id)
{
  Element * xml_classes = find_exactly_one_node<Element>(_payload, "classes");
  for (mva_class::ptrs::const_iterator cit = job.classes().begin();
      cit != job.classes().end();
      ++cit)
  {
    const string class_xpath = "class[@id='" + lexical_cast<string>(cit->id()) + "']";
    Element * xml_class = find_exactly_one_node<Element>(xml_classes, class_xpath);

    // We need throughput and rtime of last calculated host, so overwriting 
    // its value for each host is fine.
    xml_class->set_attribute("throughput", number_to_string(cit->throughput()));
    xml_class->set_attribute("rtime", number_to_string(cit->residence_time()));

    const string devices_xpath = "host[@id='" + lexical_cast<string>(host_id) + "']";
    Element * xml_devices = find_exactly_one_node<Element>(xml_class, devices_xpath);

    for (mva_device::ptrs::const_iterator dit = cit->devices().begin();
        dit != cit->devices().end();
        ++dit)
    {
      if (dit->type() != mva_device::loadindep)
        continue;

      Element * xml_parent;
      if (dit->id()[0] == 'H') 
      {
        const string parent_xpath = "host[@id='" + dit->id().substr(1) + "']";
        xml_parent = find_exactly_one_node<Element>(xml_class, parent_xpath);
      }
      else
      {
        const string parent_xpath = "device[@id='" + dit->id() + "' and @type='loadindep']";
        xml_parent = find_exactly_one_node<Element>(xml_devices, parent_xpath);
      }

      // Overwrite if already present.
      xml_parent->set_attribute("util", number_to_string(dit->service_demand() * cit->throughput()));
      xml_parent->set_attribute("rtime", number_to_string(dit->residence_time()));
      xml_parent->set_attribute("wtime", number_to_string(dit->wtime()));
    }
  }
}

  template<typename S, typename C>
void calc_mva_base<S, C>::generate_client_payload(const mva_job & client_job, unsigned client_host_id)
{
  Element * xml_classes = find_exactly_one_node<Element>(_payload, "classes");
  for (mva_class::ptrs::const_iterator cit = client_job.classes().begin();
      cit != client_job.classes().end();
      ++cit)
  {
    const string class_xpath = "class[@id='" + cit->id() + "']";
    Element * xml_class = find_exactly_one_node<Element>(xml_classes, class_xpath);

    //xml_class->set_attribute("throughput", lexical_cast<string>(cit->throughput()));
    //xml_class->set_attribute("rtime", lexical_cast<string>(cit->residence_time()));

    const string devices_xpath = "host[@id='" + lexical_cast<string>(client_host_id) + "']";
    Element * xml_devices = find_exactly_one_node<Element>(xml_class, devices_xpath);

    // We have to increment the rtime of the class with IO/CPU-times from client.
    double rtime = get_attr_value<double>(xml_class, "rtime");

    for (mva_device::ptrs::const_iterator dit = cit->devices().begin();
        dit != cit->devices().end();
        ++dit)
    {
      rtime += dit->residence_time();

      if (dit->type() != mva_device::loadindep)
        continue;

      const string parent_xpath = "device[@id='" + dit->id() + "' and @type='loadindep']";
      Element * xml_parent = find_exactly_one_node<Element>(xml_devices, parent_xpath);

      xml_parent->set_attribute("util", number_to_string(dit->service_demand() * cit->throughput()));
      xml_parent->set_attribute("rtime", number_to_string(dit->residence_time()));
      xml_parent->set_attribute("wtime", number_to_string(dit->wtime()));
    }

    // Set new residence time for root.
    xml_class->set_attribute("rtime", number_to_string(rtime));
  }
}

  template<typename S, typename C>
void calc_mva_base<S, C>::generate_root_payload(const mva_job & job, unsigned root_host_id)
{
  for (mva_class::ptrs::const_iterator cit = job.classes().begin();
      cit != job.classes().end();
      ++cit)
  {
    Element * xml_class = find_exactly_one_node<Element>(_payload, "classes/class[@id='" + cit->id() + "']"); 

    mva_device::ptrs::const_iterator dit = find_if(cit->devices().begin(), cit->devices().end(), 
        bind(&mva_device::id, _1) == "DEV" && bind(&mva_device::type, _1) == mva_device::loadindep);
    CGI_ENFORCE(dit != cit->devices().end(), "ERROR_INTERNAL_SERVER")("Could not find device with id: DEV");

    double rtime = get_attr_value<double>(xml_class, "rtime");
    rtime += dit->wtime();
    xml_class->set_attribute("rtime", number_to_string(rtime));

    Element * xml_devices = find_exactly_one_node<Element>(xml_class, "host[@id='" + lexical_cast<string>(root_host_id) + "']");
    xml_devices->set_attribute("util", number_to_string(dit->service_demand() * cit->throughput()));
    xml_devices->set_attribute("rtime", number_to_string(dit->residence_time()));
    xml_devices->set_attribute("wtime", number_to_string(dit->wtime()));
  }
}

  template<typename S, typename C>
mva_job calc_mva_base<S, C>::do_tandem_mva(unsigned host_id, 
    const vector<double> & num_customer_vector,
    unsigned indent) 
{
  _debug_control_flow << setfill(control_flow_char) << setw(indent) << "" << " START TANDEM H" << host_id << "\n";
  _debug << setfill(control_flow_char) << setw(indent) << "" << " START TANDEM H" << host_id << "\n";

  //cerr << setfill(control_flow_char) << setw(indent) << "" << " START TANDEM H" << host_id << "\n";

  const string xpath = "softrcs/softrc[@hostId='" + lexical_cast<string>(host_id) + "']";
  const bool has_softrc = !_parms->find(xpath).empty();
  CGI_ENFORCE(has_softrc, "ERROR_INTERNAL_SERVER")("No softrcs found in tandem MVA: ")(xpath);
  const unsigned num_classes = _parms->find("classes/class").size();
  const double epsilon = 0.01;
  double max_b = .0;
  vector<double> prev_blocking_proc_vector(num_classes, .0);
  bool forced_tandem_break = false;
  unsigned sqn_hqn_iterations = 0;
  sqn_mva_result result;

  mva_job hqn_job;
  do
  {
    ++sqn_hqn_iterations;

    /*
       if (sqn_hqn_iterations == 1) 
       {
    // We need to get the main-threads into the SQN calculation, which is calculated by HQN.
    const vector<double> zero_blocking_proc_vector(num_classes, .0);
    hqn_job = do_hardware_mva(host_id, zero_blocking_proc_vector, indent+control_flow_indent);
    }
     */

    // Calculate SQN (hardware job is empty in 1st sqn_hqn_iterations -> no adjustments are done).
    result = do_sqn_mva(host_id, hqn_job, num_customer_vector, prev_blocking_proc_vector, indent+control_flow_indent);

    // Calculate HQN with blocking process vector for adjustment of customer vector.
    hqn_job = do_hqn_mva(host_id, num_customer_vector, result._blocking_proc_vector, indent+control_flow_indent);

    // Calculate max b for convergence test.
    ostringstream prevstr, currstr, maxstr;
    for (size_t i=0; i<num_classes; ++i) 
    {
      prevstr << setprecision(6) << right << setw(13) << fixed << prev_blocking_proc_vector[i];
      currstr << setprecision(6) << right << setw(13) << fixed << result._blocking_proc_vector[i];
    }

    maxstr << fixed << setprecision(6) << result._max_b << " (epsilon=" << epsilon << ")\n";

    _debug << "\nPrevious B-vector: " << prevstr.str() << "\n";
    _debug << "Current B-vector: " << currstr.str() << "\n";
    _debug << "Max: " << maxstr.str() << "\n";

    /*
       if (host_id == 553)
       {
       cerr << prevstr.str() << "\n";
       cerr << currstr.str() << "\n";
       cerr << maxstr.str() << "\n";
       }
     */

    // Keep blocking processes vector for convergence test in next sqn_hqn_iterations.
    prev_blocking_proc_vector = result._blocking_proc_vector;
  }
  while(max_b > epsilon && sqn_hqn_iterations < max_sqn_hqn_iterations);

  if (max_b > epsilon && sqn_hqn_iterations == max_sqn_hqn_iterations) 
  {
    forced_tandem_break = true;
    _one_or_more_forced_tandem_breaks = true;
  }

  _debug_control_flow << setfill(control_flow_char) << setw(indent) << "" << " END TANDEM H" << host_id 
    << " [niter=" << sqn_hqn_iterations << "] " << (forced_tandem_break ? " *** FORCED BREAK ***\n" : "\n");
  _debug << setfill(control_flow_char) << setw(indent) << "" << " END TANDEM H" << host_id 
    << " [niter=" << sqn_hqn_iterations << "] " << (forced_tandem_break ? " *** FORCED BREAK ***\n" : "\n");

  //cerr << setfill(control_flow_char) << setw(indent) << "" << " END TANDEM H" << host_id << "\n";

  // Return the SQN job instead of HQN job so that we get the residence times including
  // the software resources.
  return result._sqn_job; 
}

  template<typename S, typename C>
mva_job calc_mva_base<S, C>::do_client_mva(const unsigned client_host_id, const unsigned indent)
{  
  _debug_control_flow << setfill(control_flow_char) << setw(indent) << "" << " START HQN CLIENT H" << client_host_id << "\n";
  _debug << setfill(control_flow_char) << setw(indent) << "" << " START HQN CLIENT H" << client_host_id << "\n";

  mva_job client_job(client_host_id, 0, 0);

  NodeSet xml_classes = _parms->find("classes/class");
  for (NodeSet::const_iterator cit = xml_classes.begin();
      cit != xml_classes.end();
      ++cit)
  {
    const string class_id = get_attr_value<string>(*cit, "id");
    const double think_time = get_attr_value<double>(*cit, "tt");
    bool is_noise = false; // ignore noise volume for now because we have 1 user only. // get_attr_value<bool>(*cit, "noise");
    double num_customers = 1.;

    client_job.classes().push_back(new mva_class(class_id, think_time, num_customers, is_noise));

    string devs_xpath = "host[@id='" + lexical_cast<string>(client_host_id) + "']";
    NodeSet xml_devices = (*cit)->find(devs_xpath + "/device");
    for (NodeSet::const_iterator dit = xml_devices.begin();
        dit != xml_devices.end();
        ++dit)
    {
      const string device_id = get_attr_value(*dit, "id");
      const mva_device::type_t type = mva_device::string_to_type(get_attr_value(*dit, "type"));

      // Configure device from the average tranaction (GUI), but not the loaddependent device (root).
      if (type == mva_device::delay || type == mva_device::loadindep) 
      {
        const double service_time_per_visit = get_attr_value<double>(*dit, "si");
        const double visits_per_customer = get_attr_value<double>(*dit, "vi");
        const unsigned residence_time_multiplier = get_attr_value<double>(*dit, "rmul");
        const unsigned parallelism = get_attr_value<unsigned>(*dit, "par");

        client_job.classes().back().devices().push_back(new mva_device(
              device_id, 
              type, 
              service_time_per_visit, 
              visits_per_customer, 
              residence_time_multiplier, 
              parallelism));
      }
    }
  }

  Calculator::exec(client_job);
  generate_client_payload(client_job, client_host_id);

  _debug << "\nHQN report H" << client_job.host_id() << " [CLIENT]:\n";
  client_job.print_report(_debug);
  _debug << "\n";
  _debug_control_flow << setfill(control_flow_char) << setw(indent) << "" << " END HQN CLIENT H" << client_host_id << "\n";
  _debug << setfill(control_flow_char) << setw(indent) << "" << " END HQN CLIENT H" << client_host_id << "\n";

  return client_job;
}

  template<typename S, typename C>
void calc_mva_base<S, C>::do_root_mva(unsigned root_host_id, 
    const mva_job & main_job, 
    const mva_job & client_job, 
    const unsigned indent) 
{
  _debug_control_flow << setfill(control_flow_char) << setw(indent) << "" << " START HQN WEB H" << root_host_id << "\n";
  _debug << setfill(control_flow_char) << setw(indent) << "" << " START HQN WEB H" << root_host_id << "\n";

  const Element * xml_devices = find_exactly_one_node<Element>(_parms, "classes/class[1]/host[@id=" + 
      lexical_cast<string>(root_host_id) + "]");
  const unsigned threads = get_attr_value<unsigned>(xml_devices, "threads");
  const unsigned cluster_size = get_attr_value<unsigned>(xml_devices, "clusterSize");
  mva_job root_job(root_host_id, 0, 0);

  NodeSet xml_classes = _parms->find("classes/class");
  for (NodeSet::const_iterator cit = xml_classes.begin();
      cit != xml_classes.end();
      ++cit)
  {
    const string class_id = get_attr_value<string>(*cit, "id");
    const double think_time = get_attr_value<double>(*cit, "tt");
    bool is_noise = get_attr_value<bool>(*cit, "noise");
    double num_customers = max<double>(1., get_attr_value<double>(*cit, "n") / cluster_size);

    mva_class::ptrs::const_iterator main_mva_cit = find_if(main_job.classes().begin(), main_job.classes().end(), 
        bind(&mva_class::id, _1) == class_id);
    CGI_ENFORCE(main_mva_cit != main_job.classes().end(), "ERROR_INTERNAL_SERVER")
      ("Could not find class with id: ")(class_id);

    if (is_noise)
    {
      // We cannot have a noise volume because there is no CPU device to test the
      // util% against. 
      if (Sub::identifier() == "calc-exact-mva")
      {
        // We just reset the number of customers to the value
        // the main root_job ended.
        if (!main_job.customer_vector_ended()) // If true -> j == 0.
          num_customers = main_mva_cit->j();
      }
      else if (Sub::identifier() == "calc-approx-mva")
      {
        // This is quick-n-dirty because num_customers is recalculated in the
        // innner algorithm implying that this solution depends on the 
        // internal implementation of the MVA approx alogorithm.
        num_customers = main_mva_cit->num_customers();
      }
      else
      {
        CGI_ENFORCE(false, "ERROR_INTERNAL_SERVER")("Unknown identifier: ")(Sub::identifier());
      }

      is_noise = false;
    }

    root_job.classes().push_back(new mva_class(class_id, think_time, num_customers, is_noise));

    double threads_factor = 1.0 / threads;
    root_job.classes().back().devices().push_back(new mva_device(
          "DEV", 
          mva_device::loadindep, 
          threads_factor * main_mva_cit->residence_time(), // service time per visit
          1, // visits per customer
          1, // residence time multiplier 
          1)); // parallelism

    threads_factor = ((double) threads - 1.0) / threads;
    root_job.classes().back().devices().push_back(new mva_device(
          "DEV", 
          mva_device::delay,
          threads_factor * main_mva_cit->residence_time(), // service time per visit
          1, // visits per customer
          1, // residence time multiplier 
          1)); // parallelism

    if (!client_job.classes().empty()) // No configured client.
    {
      // Configure the client as a delay device.
      // Use residence time from client MVA.

      mva_class::ptrs::const_iterator client_mva_cit = find_if(client_job.classes().begin(), client_job.classes().end(), 
          bind(&mva_class::id, _1) == class_id);
      CGI_ENFORCE(client_mva_cit != client_job.classes().end(), "ERROR_INTERNAL_SERVER")
        ("Could not find class with id: ")(class_id);

      root_job.classes().back().devices().push_back(new mva_device(
            "CLT", 
            mva_device::delay,
            client_mva_cit->residence_time(), // service time per visit
            1, // visits per customer
            1, // residence time multiplier 
            1)); // parallelism
    }
  }

  Calculator::exec(root_job);
  generate_root_payload(root_job, root_host_id);

  _debug << "\nHQN report H" << root_job.host_id() << " [WEB]:\n";
  root_job.print_report(_debug);
  _debug << "\n";
  _debug_control_flow << setfill(control_flow_char) << setw(indent) << "" << " END HQN WEB H" << root_host_id << "\n";
  _debug << setfill(control_flow_char) << setw(indent) << "" << " END HQN WEB H" << root_host_id << "\n";
}

  template<typename S, typename C>
mva_job calc_mva_base<S, C>::do_mva(unsigned host_id, 
    const vector<double> & num_customer_vector,
    unsigned indent)
{
  const bool has_softrc = !_parms->find("softrcs/softrc[@hostId='" + lexical_cast<string>(host_id) + "']").empty();
  const unsigned num_classes = _parms->find("classes/class").size();
  vector<double> blocking_vector(num_classes, .0);
  mva_job result;

  if (has_softrc)
  {
    result = do_tandem_mva(host_id, num_customer_vector, indent);
  }
  else
  {
    result = do_hqn_mva(host_id, num_customer_vector, blocking_vector, indent);
  }

  return result;
}

  template<typename S, typename C>
sqn_mva_result calc_mva_base<S, C>::do_sqn_mva(const unsigned local_host_id, 
    const mva_job & hqn_job, 
    const vector<double> & num_customer_vector, 
    const vector<double> & prev_blocking_proc_vector, 
    unsigned indent) 
{  
  const unsigned num_classes = _parms->find("classes/class").size();
  sqn_mva_result result;
  result._blocking_proc_vector = std::vector<double>(num_classes, .0);

  const NodeSet xml_softrcs = _parms->find("softrcs/softrc[@hostId='" + lexical_cast<string>(local_host_id) + "']");
  for (NodeSet::const_iterator sit = xml_softrcs.begin();
      sit != xml_softrcs.end();
      ++sit)
  {
    result = do_sqn_mva_item(*sit, sit == xml_softrcs.begin(),
        local_host_id, 
        hqn_job, 
        num_customer_vector, 
        result._blocking_proc_vector, 
        indent);
  }

  return result;
}

  template<typename S, typename C>
sqn_mva_result calc_mva_base<S, C>::do_sqn_mva_item(const Node * xml_softrc,
    const bool first_softrc,
    const unsigned local_host_id, 
    const mva_job & hqn_job, 
    const vector<double> & num_customer_vector, 
    const vector<double> & prev_blocking_proc_vector, 
    unsigned indent) 
{
  const string softrc_id = get_attr_value(xml_softrc, "id");

  _debug_control_flow << setfill(control_flow_char) << setw(indent) << "" << " START SQN H" << local_host_id << " " << softrc_id << "\n";
  _debug << setfill(control_flow_char) << setw(indent) << "" << " START SQN H" << local_host_id << "\n";

  //cerr << setfill(control_flow_char) << setw(indent) << "" << " START SQN H" << local_host_id << "\n";

  Element * xml_devices = find_exactly_one_node<Element>(_parms, "classes/class[1]/host[@id=" + 
      lexical_cast<string>(local_host_id) + "]");
  const unsigned host_cluster_size = get_attr_value<unsigned>(xml_devices, "clusterSize");
  const double servers = get_attr_value<double>(xml_softrc, "servers") / host_cluster_size;

  mva_job job(local_host_id, 0, 0), dep_job;
  map<unsigned, mva_job> tier2job; // Store recurive results per host id.

  unsigned class_index = 0;
  const NodeSet xml_classes = _parms->find("classes/class[@noise='0']");
  for (NodeSet::const_iterator cit = xml_classes.begin();
      cit != xml_classes.end();
      ++cit, ++class_index)
  {
    const string class_id = get_attr_value<string>(*cit, "id");
    double think_time = get_attr_value<double>(*cit, "tt");

    // It is important that we use the num_customers and do *not* substract the blocking customers.
    // See page 443 step 2.
    const double num_customers = max<double>(1., num_customer_vector[class_index] / host_cluster_size);

    double total_assigned_cpu = .0;
    double total_assigned_io = .0;
    double total_assigned_net = .0;
    map<const unsigned, double> total_assigned_tiers;

    job.classes().push_back(new mva_class(class_id, think_time, num_customers, false));

    const NodeSet xml_softrc_classes = xml_softrc->find("classes/class[@id='" + class_id + "']");
    CGI_ENFORCE(xml_softrc_classes.size() <= 1, "ERROR_INTERNAL_SERVER")("Volume ")(class_id)(" added twice for softrc.");
    if (xml_softrc_classes.empty()) // This softrc not active for current class volume.
    {
      job.classes().back().devices().push_back(new mva_device(
            softrc_id, 
            mva_device::loadindep, 
            0, // service time per visit
            0, // visits per customer
            1, // residence time multiplier 
            1)); // parallelism

      job.classes().back().devices().push_back(new mva_device(
            softrc_id, 
            mva_device::delay,
            0, // service time per visit
            0, // visits per customer
            1, // residence time multiplier 
            1)); // parallelism

      continue;
    }

    /*
     * At this point we have a software resource that is active for the current class and host.
     *
     * In case of loadinped devices, we have to assign a portion of the selected service demands
     * to the software resource device (delay and loadindep). For now we divide
     * the service demand equally among the number of software resources competing for the device.
     * Later, the portion is explicitely set by the user in the GUI.
     *
     * In case of loaddep devices, for now they are implicitely selected for all active software 
     * resources for the current class and therefor the portion is equally divided over all 
     * software resources.
     */

    const Node * xml_softrc_class = dynamic_cast<const Node*>(*xml_softrc_classes.begin());
    const double assigned_cpu = get_attr_value<double>(xml_softrc_class, "cpu");
    const double assigned_io = get_attr_value<double>(xml_softrc_class, "io");
    const double assigned_net = get_attr_value<double>(xml_softrc_class, "net"); // This is always 1 (=100%).
    map<const unsigned, const double> assigned_tiers;
    const double softrc_think_time = get_attr_value<double>(xml_softrc, "tt");
    const double disturbance = get_attr_value<double>(xml_softrc, "dist");
    double service_demand_cs = 0; //softrc_think_time + disturbance;

    total_assigned_cpu += assigned_cpu;
    total_assigned_io += assigned_io;
    total_assigned_net += assigned_net;

    NodeSet xml_tiers = xml_softrc_class->find("tier");
    for (NodeSet::const_iterator tit = xml_tiers.begin();
        tit != xml_tiers.end();
        ++tit)
    {
      const unsigned tier_host_id = get_attr_value<unsigned>(*tit, "id");
      const double assigned_tier = get_attr_value<double>(*tit, "factor");
      assigned_tiers.insert(make_pair(tier_host_id, assigned_tier));

      map<unsigned, double>::iterator ait = total_assigned_tiers.find(tier_host_id);
      if (ait == total_assigned_tiers.end()) {
        total_assigned_tiers.insert(make_pair(tier_host_id, assigned_tier));
      }
      else {
        ait->second += assigned_tier;
      }
    }

    const string devs_xpath = "host[@id='" + lexical_cast<string>(local_host_id) + "']";
    NodeSet xml_devices = (*cit)->find(devs_xpath + "/device");
    for (NodeSet::const_iterator dit = xml_devices.begin();
        dit != xml_devices.end();
        ++dit)
    {
      string device_id = get_attr_value(*dit, "id");
      mva_device::type_t device_type = mva_device::string_to_type(get_attr_value(*dit, "type"));
      double service_demand = get_attr_value<double>(*dit, "si") * get_attr_value<double>(*dit, "vi");

      if (device_id == "CPU") // delay and loadindep
      {
        if (hqn_job.host_id() == 0) {
          // We are in the initialization phase, use 15.5.31 in book on page 443.
          service_demand_cs += assigned_cpu * service_demand;
        }
        else {
          // We are NOT in the initialization phase, use formula 15.5.36 in book on page 444.
          service_demand_cs += assigned_cpu * hqn_residence_time(hqn_job, class_id, device_id, device_type);
        }
      }
      else if (device_id == "IO") // delay and loadindep
      {
        if (hqn_job.host_id() == 0) {
          // We are in the initialization phase, use 15.5.31 in book on page 443.
          service_demand_cs += assigned_io * service_demand;
        }
        else {
          // We are NOT in the initialization phase, use formula 15.5.36 in book on page 444.
          service_demand_cs += assigned_io * hqn_residence_time(hqn_job, class_id, device_id, device_type);
        }
      }
      else if (device_id.substr(0, 3) == "NO-" || // loadindep
          device_id.substr(0, 3) == "NI-" || // loadindep
          device_id.substr(0, 3) == "NL-") // delay
      {
        if (hqn_job.host_id() == 0) {
          // We are in the initialization phase, use 15.5.31 in book on page 443.
          service_demand_cs += assigned_net * service_demand;
        }
        else {
          // We are NOT in the initialization phase, use formula 15.5.36 in book on page 444.
          service_demand_cs += assigned_net * hqn_residence_time(hqn_job, class_id, device_id, device_type);
        }
      }
      else if (device_id[0] == 'H' && first_softrc) // delay and loadindep
      {
        unsigned tier_host_id = lexical_cast<unsigned>(device_id.substr(1));
        map<const unsigned, const double>::const_iterator ait = assigned_tiers.find(tier_host_id);
        CGI_ENFORCE(ait != assigned_tiers.end(), "ERROR_INTERNAL_SERVER")
          ("Could not find next tier with host id ")(tier_host_id)(" for volume ")(class_id);
        const double assigned_tier = ait->second;
        if (assigned_tier) 
        {
          if (tier2job.empty()) 
          {
            // We only have to calculate once, i.e. not for all classes.
            tier2job.insert(make_pair(tier_host_id, 
                  do_mva(tier_host_id, num_customer_vector, indent+control_flow_indent)));
          }

          const mva_job & tier_job = tier2job.find(tier_host_id)->second;
          service_demand_cs += assigned_tier * hqn_residence_time(tier_job, class_id);
        }
      }
    }

    // Quick fix: If there is no CPU, disk, net-IO enabled in GUI, 
    // regard the softrc as switched off for this volume.
    if (service_demand_cs > 0) {
      service_demand_cs += softrc_think_time + disturbance;
    }    

    double server_factor = 1.0 / servers;
    job.classes().back().devices().push_back(new mva_device(
          softrc_id, 
          mva_device::loadindep, 
          server_factor * service_demand_cs,
          1, 
          1, 
          1));

    server_factor = ((double) servers - 1.0) / servers;
    job.classes().back().devices().push_back(new mva_device(
          softrc_id, 
          mva_device::delay, 
          server_factor * service_demand_cs, 
          1,
          1, 
          1));

    CGI_ENFORCE(total_assigned_cpu <= 1, "ERROR_INTERNAL_SERVER")
      ("More than 100\% allocated CPU for class ")(class_id)(": ")(100.0 * total_assigned_cpu)(".");
    CGI_ENFORCE(total_assigned_io <= 1, "ERROR_INTERNAL_SERVER")
      ("More than 100\% allocated IO for class ")(class_id)(": ")(100.0 * total_assigned_io)(".");
    for (map<unsigned, double>::const_iterator ait = total_assigned_tiers.begin();
        ait != total_assigned_tiers.end();
        ++ait)
    {
      CGI_ENFORCE(ait->second <= 1, "ERROR_INTERNAL_SERVER")
        ("More than 100\% allocated tier for class ")(class_id)
        (" and tier with host id ")(ait->first)(": ")(100.0 * total_assigned_io)(".");
    }

    double service_demand_ncs = .0;
    for (NodeSet::const_iterator dit = xml_devices.begin();
        dit != xml_devices.end();
        ++dit)
    {
      const string device_id = get_attr_value(*dit, "id");
      const mva_device::type_t device_type = mva_device::string_to_type(get_attr_value(*dit, "type"));
      const double service_demand = get_attr_value<double>(*dit, "si") * get_attr_value<double>(*dit, "vi");

      if (device_id == "CPU") // delay and loadindep
      {
        if (hqn_job.host_id() == 0) {
          // We are in the initialization phase, use 15.5.31 in book on page 443.
          service_demand_ncs += (1.0 - total_assigned_cpu) * service_demand;
        }
        else {
          // We are NOT in the initialization phase, use formula 15.5.36 in book on page 444.
          service_demand_ncs += (1.0 - total_assigned_cpu) * hqn_residence_time(hqn_job, class_id, device_id, device_type);
        }
      }
      else if (device_id == "IO") // delay and loadindep
      {
        if (hqn_job.host_id() == 0) {
          // We are in the initialization phase, use 15.5.31 in book on page 443.
          service_demand_ncs += (1.0 - total_assigned_io) * service_demand;
        }
        else {
          // We are NOT in the initialization phase, use formula 15.5.36 in book on page 444.
          service_demand_ncs += (1.0 - total_assigned_io) * hqn_residence_time(hqn_job, class_id, device_id, device_type);
        }
      }
      else if (device_id[0] == 'H') // delay and loadindep
      {
        const unsigned tier_host_id = lexical_cast<unsigned>(device_id.substr(1));
        const map<const unsigned, double>::const_iterator ait = total_assigned_tiers.find(tier_host_id);
        CGI_ENFORCE(ait != total_assigned_tiers.end(), "ERROR_INTERNAL_SERVER")
          ("Could not find next tier with host id ")(tier_host_id)(" for volume ")(class_id);
        const double total_assigned_tier = ait->second;

        if (hqn_job.host_id() == 0) {
          // We are in the initialization phase, use 15.5.31 in book on page 443.
          service_demand_ncs += (1.0 - total_assigned_tier) * service_demand;
        }
        else {
          // We are NOT in the initialization phase, use formula 15.5.36 in book on page 444.
          // The loaddep device H___ is represented by a loadindep and delay device in HQN. 
          service_demand_ncs += (1.0 - total_assigned_tier) * (hqn_residence_time(hqn_job, class_id, device_id, mva_device::loadindep) +
              hqn_residence_time(hqn_job, class_id, device_id, mva_device::delay));
        }
      }
    }

    job.classes().back().devices().push_back(new mva_device(
          "NCS", 
          mva_device::delay, 
          service_demand_ncs, 
          1, 
          1, 
          1));
  }

  Calculator::exec(job);
  generate_payload(job, local_host_id);

  // Derive blocking processes vector from result SQN.
  const unsigned num_classes = _parms->find("classes/class").size();
  vector<double> blocking_proc_vector(num_classes, .0);
  const int blocking_proc_divider_hack = get_text<int>(_parms, "blockingProcDividerHack");
  CGI_ENFORCE(blocking_proc_divider_hack, "ERROR_INTERNAL_SERVER");
  stringstream blocking_proc_vector_str;
  const bool has_noise = !_parms->find("classes/class[@noise='1']").empty();
  double max_b = 0;
  for (unsigned i=0; i<num_classes; ++i) 
  {
    if (has_noise) 
    {
      if (i>0) 
      {
        // Divide as Michael's hack-request.
        blocking_proc_vector[i] = job.classes()[i-1].avg_num_blocked_customers() / blocking_proc_divider_hack; // Why -1???

        if (i>1) {
          blocking_proc_vector_str << ",";
        }
        blocking_proc_vector_str << fixed << setprecision(6) << blocking_proc_vector[i];
      }
    }
    else 
    {
      // Divide as Michael's hack-request.
      blocking_proc_vector[i] = job.classes()[i].avg_num_blocked_customers() / blocking_proc_divider_hack;

      if (i>0) {
        blocking_proc_vector_str << ",";
      }
      blocking_proc_vector_str << fixed << setprecision(6) << blocking_proc_vector[i];
    }

    max_b = max(max_b, std::fabs(((double) blocking_proc_vector[i] - prev_blocking_proc_vector[i]) / blocking_proc_vector[i]));
  }

  _debug_control_flow << setfill(control_flow_char) << setw(indent) << "" << " END SQN H" << local_host_id 
    << " [maxb=" << fixed << setprecision(6) << max_b 
    << " bvec=" << blocking_proc_vector_str.str() << "]\n";
  //cerr << setfill(control_flow_char) << setw(indent) << "" << " END SQN H" << local_host_id << "\n";

  _debug << "\nSQN report H" << job.host_id() << ":\n";
  job.print_report(_debug);
  _debug << "\n";
  _debug << setfill(control_flow_char) << setw(indent) << "" << " END SQN H" << local_host_id << "\n";

  const sqn_mva_result result = { blocking_proc_vector, max_b, job };
  return result;
  return result;
}

  template<typename S, typename C>
mva_job calc_mva_base<S, C>::do_hqn_mva(const unsigned host_id, 
    const vector<double> & num_customer_vector, 
    const vector<double> & blocking_proc_vector, 
    unsigned indent) 
{
  _debug_control_flow << setfill(control_flow_char) << setw(indent) << "" << " START HQN H" << host_id << "\n";
  //cerr << setfill(control_flow_char) << setw(indent) << "" << " START HQN H" << host_id << "\n";
  _debug << setfill(control_flow_char) << setw(indent) << "" << " START HQN H" << host_id << "\n";

  Element * xml_devices = find_exactly_one_node<Element>(_parms, "classes/class[1]/host[@id=" + 
      lexical_cast<string>(host_id) + "]");
  const double host_cpu_noise = get_attr_value<double>(xml_devices, "cpuNoise");
  const double host_disk_noise = get_attr_value<double>(xml_devices, "diskNoise");
  const unsigned host_cluster_size = get_attr_value<unsigned>(xml_devices, "clusterSize");

  mva_job job(host_id, host_cpu_noise, host_disk_noise);
  map<unsigned, mva_job> host2job; // Store recurive results per host id.

  unsigned class_index = 0;
  NodeSet xml_classes = _parms->find("classes/class");
  for (NodeSet::const_iterator cit = xml_classes.begin();
      cit != xml_classes.end();
      ++cit, ++class_index)
  {
    const string class_id = get_attr_value<string>(*cit, "id");
    const double think_time = get_attr_value<double>(*cit, "tt");
    const bool is_noise = get_attr_value<bool>(*cit, "noise");
    const double num_customers = max<double>(1., num_customer_vector[class_index] / host_cluster_size - 
        blocking_proc_vector[class_index]);

    job.classes().push_back(new mva_class(class_id, think_time, num_customers, is_noise));

    string devs_xpath = "host[@id='" + lexical_cast<string>(host_id) + "']";
    NodeSet xml_devices = (*cit)->find(devs_xpath + "/device");
    for (NodeSet::const_iterator dit = xml_devices.begin();
        dit != xml_devices.end();
        ++dit)
    {
      const string device_id = get_attr_value(*dit, "id");
      const mva_device::type_t type = mva_device::string_to_type(get_attr_value(*dit, "type"));

      if (type == mva_device::delay || type == mva_device::loadindep) 
      {
        double service_time_per_visit = get_attr_value<double>(*dit, "si");
        double visits_per_customer = get_attr_value<double>(*dit, "vi");
        const unsigned residence_time_multiplier = get_attr_value<double>(*dit, "rmul");
        const unsigned parallelism = get_attr_value<unsigned>(*dit, "par");

        job.classes().back().devices().push_back(new mva_device(
              device_id, 
              type, 
              service_time_per_visit,
              visits_per_customer, 
              residence_time_multiplier, 
              parallelism));
        continue;
      }

      CGI_ENFORCE(device_id[0] == 'H', "ERROR_INTERNAL_SERVER")("Expected an H-device");
      const unsigned dep_host_id = lexical_cast<unsigned>(device_id.substr(1));
      if (cit == xml_classes.begin()) // Only recurse for first class.
      {
        vector<double> new_num_customer_vector = num_customer_vector;
        for (unsigned i=0; i<num_customer_vector.size(); ++i) {
          new_num_customer_vector[i] -= host_cluster_size * blocking_proc_vector[i];
        }

        // The device id is of format "H\\d+" where \\d+ is dependent host id.
        host2job.insert(make_pair(dep_host_id, do_mva(dep_host_id, new_num_customer_vector, indent+control_flow_indent)));
      }

      // We only recurse for the first class, so for all subsequent classes we
      // have to recall the dependent job from the map.
      const mva_job & dep_job = host2job.find(dep_host_id)->second;

      mva_class::ptrs::const_iterator mva_cit = find_if(dep_job.classes().begin(), dep_job.classes().end(), 
          bind(&mva_class::id, _1) == class_id);
      CGI_ENFORCE(mva_cit != dep_job.classes().end(), "ERROR_INTERNAL_SERVER")
        ("Could not find class with id: ")(class_id);

      const string dep_devices_xpath = "host[@id='" + lexical_cast<string>(dep_host_id) + "']";
      const unsigned threads = get_attr_value<unsigned>(*cit, dep_devices_xpath, "threads");
      CGI_ENFORCE(threads > 0, "ERROR_INTERNAL_SERVER")("Found threads: ")(threads);

      job.classes().back().devices().push_back(new mva_device(
            device_id, 
            mva_device::loadindep, 
            (1.0 / threads) * mva_cit->residence_time(), // service demand
            1, // visits per customer
            1, // residence time multiplier 
            1)); // parallelism

      job.classes().back().devices().push_back(new mva_device(
            device_id, 
            mva_device::delay, 
            (((double) threads - 1.0) / threads) * mva_cit->residence_time(), // service demand 
            1, // visits per customer
            1, // residence time multiplier 
            1)); // parallelism
    }
  }

  Calculator::exec(job);
  generate_payload(job, host_id);

  _debug << "\nHQN report H" << job.host_id() << ":\n";
  job.print_report(_debug);
  _debug << "\n";
  _debug_control_flow << setfill(control_flow_char) << setw(indent) << "" << " END HQN H" << host_id << "\n";
  //cerr << setfill(control_flow_char) << setw(indent) << "" << " END HQN H" << host_id << "\n";
  _debug << setfill(control_flow_char) << setw(indent) << "" << " END HQN H" << host_id << "\n";

  return job;
}

  template<typename S, typename C>
void calc_mva_base<S, C>::do_exec()
{
  const unsigned mva_timeout = dm::get_setting<unsigned>("MVA_TIMEOUT");

  //
  // The following timed_join() implementation is quick and dirty, but it does exactly what
  // we want in the context of CGI (but not fastCGI, then we have to deal with signal!):
  //
  // "When the boost::thread object that represents a thread of execution is destroyed the thread 
  // becomes detached. Once a thread is detached, it will continue executing until the invocation 
  // of the function or callable object supplied on construction has completed, or the program is 
  // terminated."
  // 
  // http://www.boost.org/doc/libs/1_38_0/doc/html/thread/thread_management.html
  //
  const ptime start(microsec_clock::local_time());
  const time_duration delta_t = seconds(mva_timeout);
  thread mva_thread(bind(&calc_mva_base<S, C>::kickstart_mva, this));
  CGI_ENFORCE(mva_thread.joinable(), "ERROR_INTERNAL_SERVER")("MVA thread is not joinable.");
  const bool joined = mva_thread.timed_join(delta_t);

  stringstream debugstr;
  debugstr << "\n\n" << setw(35) << "" << "META DATA" << "\n\n\n";
  debugstr << "Projection: " << get_attr_value(_parms, "projection", "name") << "\n";
  debugstr << "Model: " << model() << "\n";
  debugstr << "Duration: " << (microsec_clock::local_time() - start) << "\n"; 
  debugstr << "Max SQN/HQN Iterations: " << max_sqn_hqn_iterations << "\n"; 
  debugstr << "\n";

  Document doc;
  doc.create_root_node_by_import(_parms);
  doc.get_root_node()->remove_child(find_exactly_one_node(doc.get_root_node(), "projection"));
  debugstr << "\n\n" << setw(35) << "" << "INPUT" << "\n\n\n";
  doc.write_to_stream_formatted(debugstr);
  debugstr << "\n";

  doc.create_root_node_by_import(_payload);
  debugstr << "\n\n" << setw(35) << "" << "OUTPUT" << "\n\n\n";
  doc.write_to_stream_formatted(debugstr);
  debugstr << "\n";

  debugstr << "\n" << setw(35) << "" << "RECURSION CONTROL FLOW" << "\n\n\n";
  debugstr << _debug_control_flow.str() << "\n\n\n";

  debugstr << setw(35) << "" << "SQN/HQN REPORTS" << "\n\n\n";
  debugstr << _debug.str() << "\n\n";

  Element * xml_debug = _payload->add_child("debug");
  xml_debug->set_attribute("ftb", _one_or_more_forced_tandem_breaks ? "1" : "0");
  xml_debug->add_child_text(debugstr.str());

  CGI_ENFORCE(joined, "ERROR_MVA_TIMEOUT")("MVA timeout has occured, maximum runtime is: ")(mva_timeout)(" sec.");
}

registrator<calc_mva_exact> __register_calc_mva_exact;
registrator<calc_mva_approx> __register_calc_mva_approx;

} 

