
#include "base.h"
#include "factory.h"
#include "exec.h"
#include "enforcer.h"
#include "credentials.h"
#include "xml/find.h"
#include "util/translator.h"

#include <mysql++/exceptions.h>
#include <boost/shared_ptr.hpp>
#include <boost/lexical_cast.hpp>

#include <iostream> // DEBUGGING ONLY

#include <libxml++/document.h>
#include <libxml++/nodes/textnode.h>
#include <libxml++/nodes/node.h>

using namespace xmlpp;
using namespace std;
using namespace boost;
using namespace util;

namespace cgi { namespace {

  void empty_payload(Node * payload)
  {
    Node::NodeList nodes = payload->get_children();
    for (Node::NodeList::const_iterator nit = nodes.begin();
        nit != nodes.end();
        ++nit) {
      payload->remove_child(*nit);
    }
  }
}

executor::return_t executor::operator() (Document *doc)
{
  string req_name;
  _error = false;

  // Load the credentials from the database according to the information provided
  // in the XML document
  Element *elt = find_exactly_one_node<Element>(doc->get_root_node(), "credentials");
  credentials::ptr creds(new credentials(elt));

  // Handle all requests
  NodeSet nodes = find_at_least_one_node(doc->get_root_node(), "requests/*");
  for (NodeSet::const_iterator nit = nodes.begin();
      nit != nodes.end();
      ++nit)
  {
    try
    {
      // Name of the node
      req_name = (*nit)->get_name();
      // Handler available for this kind of request?
      boost::shared_ptr<base> handler(factory_holder::instance().generate(req_name));
      CGI_ENFORCE(handler, "ERROR_INTERNAL_SERVER")("Unknown XML request: ")(req_name);

      if (_auth && !handler->authenticate(*creds))
      {
        string trans = translate("ERROR_SESSION_TOKEN_EXPIRED", creds->lang_id());
        find_exactly_one_node<Element>(*nit, "response/status")->set_child_text("tokeninvalid");
        find_exactly_one_node<Element>(*nit, "response/error/code")->set_child_text("ERROR_SESSION_TOKEN_EXPIRED");
        find_exactly_one_node<Element>(*nit, "response/error/usermsg")->set_child_text(trans);
        find_exactly_one_node<Element>(*nit, "response/error/module")->set_child_text("cgi");
        find_exactly_one_node<Element>(*nit, "response/error/techmsg")->set_child_text("Session token expired");
        return creds;
      }

      // Handler is found and correctly authenticated, can now be run
      handler->execute(doc, *nit, creds);

      // Remove parms to save bandwidth (only if successfull, otherwise keep parms for logging/debugging).
      (*nit)->remove_child(find_exactly_one_node(*nit, "parms"));

      // Set the status as succeeded and return
      find_exactly_one_node<Element>(*nit, "response/status")->set_child_text("success");
      return return_t(creds, req_name, "success");
    }
    catch (const util::runtime_error &e)
    {
      string trans = translate(e.code(), creds->lang_id());
      find_exactly_one_node<Element>(*nit, "response/error/code")->set_child_text(e.code());
      find_exactly_one_node<Element>(*nit, "response/error/module")->set_child_text(e.module());
      find_exactly_one_node<Element>(*nit, "response/error/usermsg")->set_child_text(trans);
      find_exactly_one_node<Element>(*nit, "response/error/techmsg")->set_child_text(e.what());
      find_exactly_one_node<Element>(*nit, "response/error/condition")->set_child_text(e.condition());
      find_exactly_one_node<Element>(*nit, "response/error/file")->set_child_text(e.file());
      find_exactly_one_node<Element>(*nit, "response/error/line")->set_child_text(lexical_cast<string>(e.line()));
      find_exactly_one_node<Element>(*nit, "response/status")->set_child_text("failure");

      // Note: We need the payload for debug output calcmva in case of timeout!!!
      // Cannot empty payload....
      //empty_payload(find_exactly_one_node(*nit, "response/payload"));

      (*nit)->remove_child(find_exactly_one_node(*nit, "parms"));

      if (e.code() == "ERROR_INTERNAL_SERVER")
      {
        _error = true;
        srand((unsigned) time(0)); 
        unsigned bug_nr = 1000000 + (unsigned) (10000000.0 * (rand() / (RAND_MAX + 1.0)));
        find_exactly_one_node<Element>(*nit, "response/error/number")->set_child_text(lexical_cast<string>(bug_nr));
      }
    }
    catch (const mysqlpp::BadQuery &e)
    {
      _error = true;
      string trans = translate("ERROR_INTERNAL_SERVER", creds->lang_id());
      find_exactly_one_node<Element>(*nit, "response/error/code")->set_child_text("ERROR_INTERNAL_SERVER");
      find_exactly_one_node<Element>(*nit, "response/error/module")->set_child_text("unknown");
      find_exactly_one_node<Element>(*nit, "response/error/usermsg")->set_child_text(trans);
      find_exactly_one_node<Element>(*nit, "response/error/techmsg")->set_child_text(string(e.what()) + ": " + e.query());
      find_exactly_one_node<Element>(*nit, "response/error/condition")->set_child_text("unknown");
      find_exactly_one_node<Element>(*nit, "response/error/file")->set_child_text("");
      find_exactly_one_node<Element>(*nit, "response/error/line")->set_child_text("");
      find_exactly_one_node<Element>(*nit, "response/status")->set_child_text("failure");

      empty_payload(find_exactly_one_node(*nit, "response/payload"));

      srand((unsigned) time(0)); 
      unsigned bug_nr = 1000000 + (unsigned) (10000000.0 * (rand() / (RAND_MAX + 1.0)));
      find_exactly_one_node<Element>(*nit, "response/error/number")->set_child_text(lexical_cast<string>(bug_nr));
    }
    catch (const std::exception &e)
    {
      _error = true;
      string trans = translate("ERROR_INTERNAL_SERVER", creds->lang_id());
      find_exactly_one_node<Element>(*nit, "response/error/code")->set_child_text("ERROR_INTERNAL_SERVER");
      find_exactly_one_node<Element>(*nit, "response/error/module")->set_child_text("unknown");
      find_exactly_one_node<Element>(*nit, "response/error/usermsg")->set_child_text(trans);
      find_exactly_one_node<Element>(*nit, "response/error/techmsg")->set_child_text(e.what());
      find_exactly_one_node<Element>(*nit, "response/error/condition")->set_child_text("unknown");
      find_exactly_one_node<Element>(*nit, "response/error/file")->set_child_text("");
      find_exactly_one_node<Element>(*nit, "response/error/line")->set_child_text("");
      find_exactly_one_node<Element>(*nit, "response/status")->set_child_text("failure");

      empty_payload(find_exactly_one_node(*nit, "response/payload"));

      srand((unsigned) time(0)); 
      unsigned bug_nr = 1000000 + (unsigned) (10000000.0 * (rand() / (RAND_MAX + 1.0)));
      find_exactly_one_node<Element>(*nit, "response/error/number")->set_child_text(lexical_cast<string>(bug_nr));
    }
  }

  return return_t(creds, req_name, "failure");
}
}

