
#ifndef __RHAN_FACTORY__
#define __RHAN_FACTORY__

#include <map>
#include <string>

#include "util/enforcer.h"
#include "util/singleton.h"

namespace cgi
{
	class base;

	struct generator_base
	{
		virtual ~generator_base() { }
		virtual base* generate() const = 0;
	};

	/**
	 * Used by the factory to create instances of RHAN.
	 */
	template <typename RHAN>
		struct generator : public generator_base
	{
		virtual base* generate() const { return new RHAN(); }
	};

	/**
	 * Maintains a registers of requests handlers accessible through
	 * their string ID, and generates handlers instances on demand.
	 */
	struct factory
	{
		typedef std::map<std::string, generator_base*> factory_map_type;

		/**
		 * Register a new request handler.
		 */
		bool register_generator(const std::string &id, generator_base *gen)
		{
			factory_map_type::const_iterator fit = _factory_map.find(id);
			if (fit != _factory_map.end())
				return false;
			return _factory_map.insert(std::make_pair(id, gen)).second;
		}

		/**
		 * Returns a pointer to a newly-allocated handler for handling requests
		 * of type id, if such a handler has previously been registered.
		 */
		base* generate(const std::string &id) const
		{
			factory_map_type::const_iterator fit = _factory_map.find(id);
			if (fit == _factory_map.end())
				return 0;
			return fit->second->generate();
		}

		friend class util::singleton<factory>;

		private:
		factory() { }
		factory(const factory &);

		factory_map_type _factory_map;
	};

	typedef util::singleton<factory> factory_holder;

	template <typename RHAN>
		struct registrator
		{
			registrator()
			{
				bool success = factory_holder::instance().register_generator(RHAN::identifier(), 
						new generator<RHAN>());
				ENFORCE(success)("Failed to register: ")(RHAN::identifier());
			}
		};
}

#endif

