
#include "base/factory.h"
#include "base/base.h"
#include "base/enforcer.h"
#include "util/path.h"
#include "parsers/core/types.h"
#include "parsers/core/session.h"
#include "parsers/core/msr.h"
#include "parsers/core/tri.h"
#include "parsers/core/timer.h"
#include "parsers/core/runtime_error.h"
#include "util/log.h"
#include "xml/getattr.h"
#include "xml/gettext.h"

#include <unistd.h>
#include <mysql++/exceptions.h>

#include <libxml++/document.h>
#include <libxml++/nodes/element.h>
#include <libxml++/nodes/textnode.h>
#include <libxml++/validators/schemavalidator.h>

#include <log4cxx/basicconfigurator.h>
#include <log4cxx/patternlayout.h>
#include <log4cxx/fileappender.h>

#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/filesystem/operations.hpp>

#include <set>

using namespace log4cxx;
using namespace boost::filesystem;
using namespace xmlpp;
using namespace std;
using namespace boost;
using namespace boost::posix_time;
using namespace parsers;

namespace cgi {

  class parse_session : public base
  {
    public:

      parse_session() : base() { }
      virtual ~parse_session() { }

      static std::string identifier() { return "parse-session"; }

    protected:

      virtual void do_exec();

    private:

      void setup_log();
      void send_and_remove_log();
      session_ptr handle_session(Node * xml_session, 
          const unsigned counter, 
          const bool debug_log,
          const string debug_log_path,
          const session_ptr group_s);

      string _log_filename;
  };

  void parse_session::setup_log()
  {
    // Write existing parser log to 'unique' file.
    _log_filename = INSTALL_PATH "/log/cgi_parse_" + lexical_cast<string>(getpid()) + ".log";
    BasicConfigurator::resetConfiguration();
    LayoutPtr layout(new PatternLayout("[%p] %m%n"));
    AppenderPtr appender(new FileAppender(layout, _log_filename));
    BasicConfigurator::configure(appender);
  }

  void parse_session::send_and_remove_log()
  {
    // Read contents of log file.
    ifstream file(_log_filename.c_str());
    if (!file.is_open()) {
      return;
    }

    stringstream log;
    log << file.rdbuf();

    // Add contents of log to payload.
    Element * xml_log = _payload->add_child("log");
    xml_log->add_child_text(log.str());

    // Delete log file.
    file.close();
    remove(_log_filename);
  }

  session_ptr parse_session::handle_session(Node * xml_session, 
      const unsigned counter, 
      const bool debug_log,
      const string debug_log_path,
      const session_ptr group_s)
  {
    const string name = get_attr_value<string>(xml_session, "projectName");
    if (counter != 1) {
      LOG4CXX_INFO(util::logger, "\n\n\nParsing session #" << counter << ": " << name);
    }
    else {
      LOG4CXX_INFO(util::logger, "Parsing session #" << counter << ": " << name);
    }

    Document * doc = new Document();
    doc->create_root_node_by_import(xml_session);
    Element * imported_session = doc->get_root_node();
    imported_session->set_attribute("debugLog", debug_log ? "1" : "0"); 
    imported_session->set_attribute("debugLogPath", debug_log_path); 

    Node::PrefixNsMap namespaces;
    namespaces.insert(make_pair("mbrace", "http://www.contentional.eu"));

    NodeSet xml_logs = xml_session->find("//mbrace:parserLogs/*", namespaces);
    CGI_ENFORCE(!xml_logs.empty(), "ERROR_INTERNAL_SERVER");

    set<string> data_names;
    for (NodeSet::iterator nit = xml_logs.begin();
        nit != xml_logs.end();
        ++nit)
    {
      vector<string> names;
      Element * xml_rc = dynamic_cast<Element*>(*nit);
      CGI_ENFORCE(xml_rc, "ERROR_INTERNAL_SERVER");
      TextNode * data_name_node = xml_rc->get_child_text();
      CGI_ENFORCE(data_name_node, "ERROR_INTERNAL_SERVER");
      string content = data_name_node->get_content();
      split(names, content, is_any_of(",;")); 
      data_names.insert(names.begin(), names.end());
    }

    for (set<string>::const_iterator nit = data_names.begin();
        nit != data_names.end();
        ++nit)
    {
      NodeSet xml_datas = _parms->find("mbrace:data[@name='" + *nit + "']", namespaces);
      if (xml_datas.empty()) 
      {
        Element * data = doc->get_root_node()->add_child("data");
        data->set_attribute("name", *nit);
        data->set_namespace_declaration("http://www.contentional.eu");
        continue;
      }

      CGI_ENFORCE(xml_datas.size() == 1, "ERROR_INTERNAL_SERVER")("Expected one data node ")(*nit);
      doc->get_root_node()->import_node(*xml_datas.begin(), true);
    }

    // Create the session object and parse it.
    session_ptr s = session::create_from_xml(doc);
    s->parse();

    if (group_s)
    {
      CGI_ENFORCE(group_s->_msrs.size() == 1, "ERROR_INTERNAL_SERVER");
      msr_ptr group_m = group_s->_msrs.back();
      CGI_ENFORCE(group_m->_tris.size() == 1, "ERROR_INTERNAL_SERVER");
      const tri_ptr group_i = group_m->_tris.back();

      CGI_ENFORCE(s->_msrs.size() == 1, "ERROR_INTERNAL_SERVER");
      msr_ptr m = s->_msrs.back();
      CGI_ENFORCE(m->_tris.size() == 1, "ERROR_INTERNAL_SERVER");
      tri_ptr i = m->_tris.back();

      i->_group_trs_per_sec = group_i->_trs_per_sec;
    }

    s->dbdump();

    delete doc;

    return s;
  }

  void parse_session::do_exec()
  {
    try 
    {
      setup_log();

      // Initialize timer.
      timer_holder::instance().init();

      Node::PrefixNsMap namespaces;
      namespaces.insert(make_pair("mbrace", "http://www.contentional.eu"));

      bool debug_log = 0;
      string debug_log_path;
      if (!_parms->get_children("debugLog").empty() &&
          !_parms->get_children("debugLogPath").empty())
      {
        if ((debug_log = get_text<bool>(_parms, "debugLog")))
        {
          debug_log_path = get_text<string>(_parms, "debugLogPath");

          NodeSet xml_atms = _parms->find("//mbrace:parserLogs/mbrace:atm", namespaces);
          CGI_ENFORCE(!xml_atms.empty(), "ERROR_INTERNAL_SERVER");
          Element * xml_atm = dynamic_cast<Element*>(*xml_atms.begin());

          const ptime from = time_from_string(get_attr_value<string>(xml_atm, "from"));
          const ptime to = time_from_string(get_attr_value<string>(xml_atm, "to"));

          stringstream path;
          path << debug_log_path << "/"
            << to_iso_extended_string(from.date()) << "/"
            << setfill('0') << setw(2) << from.time_of_day().hours() 
            << setfill('0') << setw(2) << from.time_of_day().minutes() << "-"
            << setfill('0') << setw(2) << to.time_of_day().hours()
            << setfill('0') << setw(2) << to.time_of_day().minutes();
          debug_log_path = path.str();
          create_directories(debug_log_path);
        }
      }

      if (debug_log)
      {
        stringstream tmp;
        tmp << debug_log_path << "/remote_request_" << setw(6) << setfill('0') << getpid() << ".xml";
        ofstream out(tmp.str().c_str());
        CGI_ENFORCE(out.is_open(), "ERROR_INTERNAL_SERVER")("Failed to open file: ")(tmp.str());
        _doc->write_to_stream(out);
      }

      unsigned counter = 0;
      const NodeSet xml_group_sessions = _parms->find("//group/mbrace:session", namespaces);
      CGI_ENFORCE(xml_group_sessions.size() <= 1, "ERROR_INTERNAL_SERVER");
      session_ptr group_s;
      if (!xml_group_sessions.empty()) 
      {
        group_s = handle_session(xml_group_sessions.back(), 
            ++counter, 
            debug_log,
            debug_log_path,
            session_ptr());
      }

      // Generate a session-document for non-group sessions.
      const Node::NodeList xml_sessions = _parms->get_children("session");
      CGI_ENFORCE(xml_sessions.size() >= 1, "ERROR_INTERNAL_SERVER");
      for (Node::NodeList::const_iterator sit = xml_sessions.begin();
          sit != xml_sessions.end();
          ++sit) 
      {
        handle_session(*sit, 
            ++counter, 
            debug_log,
            debug_log_path,
            group_s);
      }

      send_and_remove_log();
    }
    catch(const util::runtime_error & e)
    {
      send_and_remove_log();

      throw e;
    }
    catch (const mysqlpp::BadQuery & e) 
    {
      send_and_remove_log();

      stringstream msg;
      msg << "Caught mysqlpp::BadQuery: " << e.query() << "\n" << e.what();
      PARSER_THROW1(msg.str());
    }
    catch(const std::exception & e)
    {
      send_and_remove_log();

      PARSER_THROW1(e.what());
    }
  }

  registrator<parse_session> __register_parse_session;
} 

